@@ -7,12 +7,15 @@
 * Copyright (C) 2005 David LaRose, dlr@cs.cmu.edu
 * See accompanying file, LICENSE.TXT, for details.
 *
-* $Revision: 997 $
-* $Date: 2008-01-16 03:17:44 -0500 (Wed, 16 Jan 2008) $
+* $Revision: 1053 $
+* $Date: 2008-10-17 16:27:02 -0400 (Fri, 17 Oct 2008) $
 ***************************************************************************
 */
 
 #include <fstream>
+
+#include <png.h>
+
 #include <dlrCommon/byteOrder.h>
 #include <dlrComputerVision/imageIO.h>
 
@@ -62,6 +65,42 @@ namespace {
 namespace dlr {
 
   namespace computerVision {
+
+
+#if 0
+    
+    // Constructor.
+    ImageReader::
+    ImageReader()
+    {
+      // Empty.
+    }
+      
+      
+    // Destructor.
+    ImageReader::
+    ~ImageReader()
+    {
+      // Empty.
+    }
+
+
+    void
+    ImageReader::
+    readFile(std::string const& fileName)
+    {
+
+    }
+
+
+    ImageFormat
+    ImageReader::
+    getNativeFormat()
+    {
+      return GRAY8;
+    }
+
+#endif
     
     Image<GRAY8>
     readPGM8(const std::string& fileName)
@@ -475,7 +514,256 @@ namespace dlr {
       // All done!
       outputStream.close();
     }
-  
+
+
+#if HAVE_LIBPNG
+
+    Image<GRAY8>
+    readPNG8(const std::string& fileName,
+             std::string& commentString)
+    {
+      // This code is heavily in debt to example.c from the libpng 1.2.1
+      // distribition, which carries the following header comment:
+      // /* example.c - an example of using libpng
+      //  * Last changed in libpng 1.2.1 December 7, 2001.
+      //  * This file has been placed in the public domain by the authors.
+      //  * Maintained 1998-2001 Glenn Randers-Pehrson
+      //  * Maintained 1996, 1997 Andreas Dilger)
+      //  * Written 1995, 1996 Guy Eric Schalnat, Group 42, Inc.)
+      //  */
+
+      Image<GRAY8> result;
+
+      // We'll check eight bytes of magic at the beginning of the file
+      // to make sure it's actually a png image.
+      const size_t pngSignatureSize = 8;
+
+      FILE *fp = fopen(fileName.c_str(), "rb");
+      if (fp == 0) {
+        DLR_THROW(dlr::IOException, "dlr::computerVision::readPng8()",
+                  "Couldn't open input file.");
+      }
+
+      // Be sure to clean up the open file.
+      try {
+
+        // Read and check the png magic to see if we have an actual
+        // png image.
+        unsigned char header[pngSignatureSize + 1];
+        if(fread(header, 1, pngSignatureSize, fp) != pngSignatureSize) {
+          DLR_THROW(IOException, "dlr::computerVision::readPng8()",
+                    "Couldn't read png signature from file.");
+        }
+        if(png_sig_cmp(header, 0, pngSignatureSize) != 0) {
+          DLR_THROW(IOException, "dlr::computerVision::readPng8()",
+                    "Fiel doesn't seem to be a PNG image.");
+        }
+        
+        // Create and initialize the png_struct with the default stderr
+        // and longjump error functions.
+        png_structp pngPtr = png_create_read_struct(
+          PNG_LIBPNG_VER_STRING, 0, 0, 0);
+        if(pngPtr == 0) {
+          DLR_THROW(dlr::RunTimeException, "dlr::computerVision::readPng8()",
+                    "Couldn't initialize png_structp.");
+        }
+
+        // This variable is just to let us choreograph nicely with libpng
+        // cleanup functions.
+        png_infop* infoPtrPtrForCleanup = png_infopp_NULL;
+
+        // Be sure to clean up pngPtr (and later, infoPtr).
+        try {
+
+          // Allocate/initialize the memory for image information.
+          png_infop infoPtr = png_create_info_struct(pngPtr);
+          if (infoPtr == 0) {
+            DLR_THROW(RunTimeException, "dlr::computerVision::readPng8()",
+                      "Couldn't initialize png_infop.");
+          }
+          infoPtrPtrForCleanup = &infoPtr;
+       
+          // Set error handling in case libpng calls longjmp().
+          if(setjmp(png_jmpbuf(pngPtr))) {
+            DLR_THROW(dlr::IOException, "dlr::computerVision::readPng8()",
+                      "Trouble reading from file.");
+          }
+        
+          // Set up the input control.
+          png_init_io(pngPtr, fp);
+
+          // Let libpng know that we've already checked some magic.
+          png_set_sig_bytes(pngPtr, pngSignatureSize);
+          
+          // Read entire image into pngPtr.
+          png_read_png(pngPtr, infoPtr, PNG_TRANSFORM_IDENTITY, png_voidp_NULL);
+
+          // Find out about our image.
+          png_uint_32 width, height;
+          int bitDepth, colorType, interlaceType;
+          int compressionType, filterMethod;
+          png_get_IHDR(pngPtr, infoPtr, &width, &height, &bitDepth, &colorType,
+                       &interlaceType, &compressionType, &filterMethod);
+          if(bitDepth != 8) {
+            DLR_THROW(IOException, "dlr::computerVision::readPng8()",
+                      "Image file is not 8 bit.");
+          }
+          if(colorType != PNG_COLOR_TYPE_GRAY) {
+            DLR_THROW(IOException, "dlr::computerVision::readPng8()",
+                      "Can currently only handle grayscale images.");
+          }
+          if(interlaceType != PNG_INTERLACE_NONE) {
+            DLR_THROW(IOException, "dlr::computerVision::readPng8()",
+                      "Can currently only handle non-interlaced images.");
+          }
+
+          // Copy image and return.
+          result.reinit(height, width);
+          png_bytep* rowPointers = png_get_rows(pngPtr, infoPtr);
+          for(size_t rowIndex = 0; rowIndex < height; ++rowIndex) {
+            result.getRow(rowIndex).copy(rowPointers[rowIndex]);
+          }
+
+        } catch(...) {
+          png_destroy_read_struct(
+            &pngPtr, infoPtrPtrForCleanup, png_infopp_NULL);
+          throw;
+        }
+        
+        // clean up after the read, and free any memory allocated.
+        png_destroy_read_struct(&pngPtr, infoPtrPtrForCleanup, png_infopp_NULL);
+
+      } catch(...) {
+        fclose(fp);
+        throw;
+      }
+      fclose(fp);
+
+      return result;
+    }
+
+    
+    void
+    writePNG8(const std::string& fileName,
+              const Image<GRAY8>& outputImage,
+              const std::string& comment)
+    {
+      // This code is heavily in debt to example.c from the libpng 1.2.1
+      // distribition, which carries the following header comment:
+      // /* example.c - an example of using libpng
+      //  * Last changed in libpng 1.2.1 December 7, 2001.
+      //  * This file has been placed in the public domain by the authors.
+      //  * Maintained 1998-2001 Glenn Randers-Pehrson
+      //  * Maintained 1996, 1997 Andreas Dilger)
+      //  * Written 1995, 1996 Guy Eric Schalnat, Group 42, Inc.)
+      //  */
+
+      FILE* fp = fopen(fileName.c_str(), "wb");
+      if (fp == 0) {
+        DLR_THROW(dlr::IOException, "ImageIO::writePng()",
+                  "Couldn't open output file.");
+      }
+
+      // Make sure we clean up open FILE.
+      try {
+
+        // Create and initialize the png_struct with the default stderr and
+        // longjump error handler functions.
+        png_structp pngPtr = png_create_write_struct(
+          PNG_LIBPNG_VER_STRING, 0, 0, 0);
+        if (pngPtr == 0) {
+          DLR_THROW(dlr::RunTimeException, "ImageIO::writePng()",
+                    "Couldn't initialize png_structp.");
+        }
+
+        // This variable is just to let us choreograph nicely with libpng
+        // cleanup functions.
+        png_infop* infoPtrPtrForCleanup = png_infopp_NULL;
+
+        // Make sure we clean up pngPtr (and eventually infoPtr).
+        try { 
+
+          // We'll need a place to record details (colorspace, etc.)
+          // about the image.
+          png_infop infoPtr = png_create_info_struct(pngPtr);
+          if (infoPtr == 0) {
+            DLR_THROW(dlr::RunTimeException, "ImageIO::writePng()",
+                      "Couldn't initialize png_infop.");
+          }
+          infoPtrPtrForCleanup = &infoPtr;
+       
+          // Set error handling in case libpng calls longjmp().
+          if(setjmp(png_jmpbuf(pngPtr))) {
+            // If we get here, we had a problem reading the file
+            DLR_THROW(dlr::IOException, "ImageIO::writePng()",
+                      "Trouble reading from file.");
+          }
+
+          // Set up the output control.
+          png_init_io(pngPtr, fp);
+
+          // Note(xxx): Just in case we need it...
+          // png_set_compression_level(pngPtr, Z_BEST_COMPRESSION);
+
+          // Prepare for writing.
+          png_set_IHDR(
+            pngPtr, infoPtr, outputImage.columns(), outputImage.rows(),
+            sizeof(ImageFormatTraits<GRAY8>::PixelType) * 8,
+            PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE,
+            PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
+         
+     
+          // xxx Fill in png structure here.
+          png_bytep* rowPointers = (png_bytep*)png_malloc(
+            pngPtr, outputImage.rows() * png_sizeof(png_bytep));
+          if(rowPointers == 0) {
+            DLR_THROW(dlr::RunTimeException, "ImageIO::writePng()",
+                      "Couldn't allocate row pointers.");
+          }
+
+          // Be sure to free rowPointers.
+          try {
+            for(size_t rowIndex = 0; rowIndex < outputImage.rows();
+                ++rowIndex) {
+              rowPointers[rowIndex] = const_cast<unsigned char*>(
+                outputImage.getRow(rowIndex).data());
+            }
+            png_set_rows(pngPtr, infoPtr, rowPointers);
+            png_write_png(
+              pngPtr, infoPtr, PNG_TRANSFORM_IDENTITY, png_voidp_NULL);
+          } catch(...) {
+            png_free(pngPtr, rowPointers);
+          }
+          png_free(pngPtr, rowPointers);
+
+          /* If you png_malloced a palette, free it here (don't free
+             infoPtr->palette, as recommended in versions 1.0.5m and
+             earlier of this example; if libpng mallocs infoPtr->palette,
+             libpng will free it).  If you allocated it with malloc()
+             instead of png_malloc(), use free() instead of
+             png_free(). */
+          // png_free(pngPtr, palette); xxx;
+          // palette=NULL;
+
+        } catch(...) {
+          png_destroy_write_struct(&pngPtr, infoPtrPtrForCleanup);
+          throw;
+        }
+       
+        /* clean up after the write, and free any memory allocated */
+        png_destroy_write_struct(&pngPtr, infoPtrPtrForCleanup);
+
+      } catch(...) {
+        fclose(fp);
+        throw;
+      }
+
+      fclose(fp);
+
+    }
+
+#endif /* #if HAVE_LIBPNG */
+    
   } // namespace computerVision    
 
 } // namespace dlr
