@@ -7,8 +7,8 @@
 * Copyright (C) 2006 David LaRose, dlr@cs.cmu.edu
 * See accompanying file, LICENSE.TXT, for details.
 *
-* $Revision: 946 $
-* $Date: 2007-07-09 20:24:50 -0400 (Mon, 09 Jul 2007) $
+* $Revision: 1155 $
+* $Date: 2009-05-24 00:07:38 -0400 (Sun, 24 May 2009) $
 ***************************************************************************
 */
 
@@ -209,20 +209,22 @@ namespace dlr {
 
     /**
      * This function subsamples its input to create a new, smaller
-     * image.  The number of rows in the resulting image is the largest
-     * integer, r, such that rowStep * r <= inputImage.rows().  The
-     * number of columns in the resulting image is the largest integer,
-     * c, such that columnStep * c <= inputImage.columns().  The pixel
-     * values in the resulting image are the values at the intersection
-     * of every rowStep-th row and columnStep-th column of inputImage,
-     * starting with the value at (row, column) = (0, 0), which is
-     * always copied directly in to pixel (0, 0) of the result image.
-     * In other words, element (0, 0) of the resulting image is equal to
+     * image.  The number of rows in the resulting image is the
+     * largest integer, r, such that (rowStep * (r - 1) + 1) <=
+     * inputImage.rows().  The number of columns in the resulting
+     * image is the largest integer, c, such that (columnStep * (c -
+     * 1) + 1) <= inputImage.columns().  The pixel values in the
+     * resulting image are the values at the intersection of every
+     * rowStep-th row and columnStep-th column of inputImage, starting
+     * with the value at (row, column) = (0, 0), which is always
+     * copied directly in to pixel (0, 0) of the result image.  In
+     * other words, element (0, 0) of the resulting image is equal to
      * element (0, 0) of inputImage, element (0, 1) of the resulting
-     * image is equal to element (0, columnStep) of inputImage, element
-     * (0, 2) of the resulting image is equal to element (0, 2 *
-     * columnStep) of inputImage, element (2, 4) of the resulting image
-     * is equal to element (2 * rowStep, 4 * columnStep) in inputImage.
+     * image is equal to element (0, columnStep) of inputImage,
+     * element (0, 2) of the resulting image is equal to element (0, 2
+     * * columnStep) of inputImage, element (2, 4) of the resulting
+     * image is equal to element (2 * rowStep, 4 * columnStep) in
+     * inputImage.
      *
      * @param inputImage This argument is the image to be subsampled.
      *
@@ -241,19 +243,23 @@ namespace dlr {
               size_t columnStep = 2);
 
 
-#if 0
     /**
      * This function interpolates its input to create a new, larger
-     * image.  The number of rows in the resulting image is twice the
-     * number of rows in the input image.  The number of columns in the
-     * resulting image is twice the number of columns in the input
-     * image.  Where output pixels overlap input pixels, they are copied
-     * directly from the input image.  Where output pixels fall in
-     * between input pixels, they are constructed by bilinear
-     * interpolation.  The pixel at (0, 0) in the output image falls
-     * directly on top of pixel (0, 0) in the input iamage.  (2, 2) in
-     * the output image falls on top of (1, 1) in the input image, and
-     * so forth.
+     * image.  The number of rows in the resulting image is one fewer
+     * than twice the number of rows in the input image.  The number
+     * of columns in the resulting image is one fewer than twice the
+     * number of columns in the input image.  Where output pixels
+     * overlap input pixels, they are copied directly from the input
+     * image.  Where output pixels fall in between input pixels, they
+     * are constructed by bilinear interpolation.  The pixel at (0, 0)
+     * in the output image falls directly on top of pixel (0, 0) in
+     * the input image.  (2, 2) in the output image falls on top of
+     * (1, 1) in the input image, and so forth.
+     *
+     * Template argument IntermediateFormat specifies the type of
+     * pixel that should be used during calculation of the
+     * interpolated values.  It should have sufficient precision to
+     * contain the sum of four input pixels.
      *
      * @param inputImage This argument is the image to be interpolated.
      *
@@ -263,10 +269,7 @@ namespace dlr {
              ImageFormat OutputFormat,
              ImageFormat IntermediateFormat>
     Image<OutputFormat>
-    supersample(const Image<InputFormat>& inputImage,
-                typename ImageFormatTraits<OutputFormat>::PixelType,
-                typename ImageFormatTraits<IntermediateFormat>::PixelType);
-#endif
+    supersample(const Image<InputFormat>& inputImage);
   
     
     /** 
@@ -655,8 +658,21 @@ namespace dlr {
               size_t rowStep,
               size_t columnStep)
     {
-      Image<Format> outputImage(inputImage.rows() / rowStep,
-                                inputImage.columns() / columnStep);
+      // Argument checking.
+      if(inputImage.size() == 0) {
+        return Image<Format>();
+      }
+      if(rowStep == 0) {
+        rowStep = 1;
+      }
+      if(columnStep == 0) {
+        columnStep = 1;
+      }
+
+      // Output image size is at least (1, 1).  All rows and columns
+      // must map to valid positions in the input image.
+      Image<Format> outputImage((inputImage.rows() - 1) / rowStep + 1,
+                                (inputImage.columns() - 1) / columnStep + 1);
       for(size_t outputRowIndex = 0; outputRowIndex < outputImage.rows();
           ++outputRowIndex) {
         Array1D<typename Image<Format>::PixelType> inputRow =
@@ -675,60 +691,93 @@ namespace dlr {
     }
 
 
-#if 0
     // This function interpolates its input to create a new, larger
     // image.
     template<ImageFormat InputFormat,
              ImageFormat OutputFormat,
              ImageFormat IntermediateFormat>
     Image<OutputFormat>
-    supersample(const Image<InputFormat>& inputImage,
-                typename ImageFormatTraits<OutputFormat>::PixelType,
-                typename ImageFormatTraits<IntermediateFormat>::PixelType)
+    supersample(const Image<InputFormat>& inputImage)
     {
-      Image<OutputFormat> outputImage(inputImage.rows() * 2,
-                                      inputImage.columns() * 2);
-      for(size_t inputRowIndex = 0; inputRowIndex < inputImage.rows();
+      Image<OutputFormat> outputImage(inputImage.rows() * 2 - 1,
+                                      inputImage.columns() * 2 - 1);
+      ColorspaceConverter<InputFormat, OutputFormat> converter0;
+      ColorspaceConverter<InputFormat, IntermediateFormat> converter1;
+      ColorspaceConverter<IntermediateFormat, OutputFormat> converter2;
+
+      // Working variables.
+      typename ImageFormatTraits<IntermediateFormat>::PixelType tempPixel;
+      Array1D<typename Image<InputFormat>::PixelType> inputRow0;
+      Array1D<typename Image<InputFormat>::PixelType> inputRow1;
+      Array1D<typename Image<OutputFormat>::PixelType> outputRow0;
+      Array1D<typename Image<OutputFormat>::PixelType> outputRow1;
+
+      // The main supersampling loop.
+      for(size_t inputRowIndex = 0; inputRowIndex < (inputImage.rows() - 1);
           ++inputRowIndex) {
-        Array1D<typename Image<InputFormat>::PixelType> inputRow0 =
-          inputImage.row(inputRowIndex);
-        Array1D<typename Image<InputFormat>::PixelType> inputRow1 =
-          inputImage.row(inputRowIndex + 1);
-        Array1D<typename Image<OutputFormat>::PixelType> outputRow0 =
-          outputImage.row(2 * inputRowIndex);
-        Array1D<typename Image<OutputFormat>::PixelType> outputRow1 =
-          outputImage.row(2 * inputRowIndex + 1);
+        inputRow0 = inputImage.row(inputRowIndex);
+        inputRow1 = inputImage.row(inputRowIndex + 1);
+        outputRow0 = outputImage.row(2 * inputRowIndex);
+        outputRow1 = outputImage.row(2 * inputRowIndex + 1);
         size_t outputColumnIndex = 0;
         for(size_t inputColumnIndex = 0;
-            inputColumnIndex < inputImage.columns();
+            inputColumnIndex < (inputImage.columns() - 1);
             ++inputColumnIndex) {
-          typename ImageFormatTraits<IntermediateFormat>::PixelType tempPixel;
-        
-          outputRow0[outputColumnIndex] = inputRow0[inputColumnIndex];
-        
-          tempPixel = inputRow0[inputColumnIndex];
-          tempPixel += inputRow0[inputColumnIndex + 1];
-          tempPixel /= 2;
-          outputRow0[outputColumnIndex + 1] = tempPixel;
+          outputRow0[outputColumnIndex] =
+            converter0(inputRow0[inputColumnIndex]);
 
-          tempPixel = inputRow0[inputColumnIndex];
-          tempPixel += inputRow1[inputColumnIndex];
+          // Note(xxx): This can be made more efficient.
+          tempPixel = converter1(inputRow0[inputColumnIndex]);
+          tempPixel += converter1(inputRow0[inputColumnIndex + 1]);
           tempPixel /= 2;
-          outputRow1[outputColumnIndex] = tempPixel;
+          outputRow0[outputColumnIndex + 1] = converter2(tempPixel);
 
-          tempPixel = inputRow0[inputColumnIndex];
-          tempPixel += inputRow0[inputColumnIndex + 1];
-          tempPixel += inputRow1[inputColumnIndex];
-          tempPixel += inputRow1[inputColumnIndex + 1];
+          tempPixel = converter1(inputRow0[inputColumnIndex]);
+          tempPixel += converter1(inputRow1[inputColumnIndex]);
+          tempPixel /= 2;
+          outputRow1[outputColumnIndex] = converter2(tempPixel);
+
+          tempPixel = converter1(inputRow0[inputColumnIndex]);
+          tempPixel += converter1(inputRow0[inputColumnIndex + 1]);
+          tempPixel += converter1(inputRow1[inputColumnIndex]);
+          tempPixel += converter1(inputRow1[inputColumnIndex + 1]);
           tempPixel /= 4;
-          outputRow1[outputColumnIndex + 1] = tempPixel;
+          outputRow1[outputColumnIndex + 1] = converter2(tempPixel);
 
           outputColumnIndex += 2;
         }
+        // Handle pixels at end of rows.
+        outputRow0[outputColumnIndex] =
+          converter0(inputRow0[inputRow0.size() - 1]);
+
+        tempPixel = converter1(inputRow0[inputRow0.size() - 1]);
+        tempPixel += converter1(inputRow1[inputRow0.size() - 1]);
+        tempPixel /= 2;
+        outputRow1[outputColumnIndex] = converter2(tempPixel);
       }
+
+      // Handle final row.
+      inputRow0 = inputImage.row(inputImage.rows() - 1);
+      outputRow0 = outputImage.row(outputImage.rows() - 1);
+      size_t outputColumnIndex = 0;
+      for(size_t inputColumnIndex = 0;
+          inputColumnIndex < (inputImage.columns() - 1);
+          ++inputColumnIndex) {
+        outputRow0[outputColumnIndex] =
+          converter0(inputRow0[inputColumnIndex]);
+
+        tempPixel = converter1(inputRow0[inputColumnIndex]);
+        tempPixel += converter1(inputRow0[inputColumnIndex + 1]);
+        tempPixel /= 2;
+        outputRow0[outputColumnIndex + 1] = converter2(tempPixel);
+
+        outputColumnIndex += 2;
+      }
+      outputRow0[outputImage.columns() - 1] =
+        inputRow0[inputImage.columns() - 1];
+
       return outputImage;
     }
-#endif
   
 
     // This function creates a new array and copies into it the pixel
